@@ -6,6 +6,7 @@ use crate::ast::analyze_dependencies;
 use crate::ast::analyze_jsdoc_imports;
 use crate::ast::analyze_jsx_import_sources;
 use crate::ast::analyze_ts_references;
+use crate::ast::DependencyKind;
 use crate::ast::SourceParser;
 use crate::module_specifier::resolve_import;
 use crate::module_specifier::ModuleSpecifier;
@@ -17,6 +18,8 @@ use deno_ast::MediaType;
 use deno_ast::ParsedSource;
 use futures::stream::FuturesUnordered;
 use futures::stream::StreamExt;
+use futures::Future;
+use futures::FutureExt;
 use serde::ser::SerializeMap;
 use serde::ser::SerializeSeq;
 use serde::ser::SerializeStruct;
@@ -30,6 +33,7 @@ use std::collections::BTreeMap;
 use std::collections::HashMap;
 use std::collections::HashSet;
 use std::fmt;
+use std::pin::Pin;
 use std::rc::Rc;
 use std::sync::Arc;
 
@@ -1306,7 +1310,14 @@ pub(crate) fn parse_module_from_ast(
       ),
       maybe_resolver,
     );
-    dep.maybe_code = resolved_dependency;
+    if matches!(
+      desc.kind,
+      DependencyKind::ImportType | DependencyKind::ExportType
+    ) {
+      dep.maybe_type = resolved_dependency;
+    } else {
+      dep.maybe_code = resolved_dependency;
+    }
     let specifier = module.specifier.clone();
     let maybe_type = analyze_deno_types(&desc)
       .map(|(text, span)| {
@@ -1350,15 +1361,33 @@ fn is_untyped(media_type: &MediaType) -> bool {
   )
 }
 
+/// The kind of build to perform.
+pub(crate) enum BuildKind {
+  /// All types of dependencies should be analyzed and included in the graph.
+  All,
+  /// Only code dependencies should be analyzed and included in the graph. This
+  /// is useful when transpiling and running code, but not caring about type
+  /// only dependnecies.
+  CodeOnly,
+  /// Only type dependencies should be analyzed and included in the graph. This
+  /// is useful when assessing types, like documentation or type checking, when
+  /// the code will not be executed.
+  TypesOnly,
+}
+
+pub type LoadWithSpecifierFuture =
+  Pin<Box<dyn Future<Output = (ModuleSpecifier, LoadResult)> + 'static>>;
+
 pub(crate) struct Builder<'a> {
   in_dynamic_branch: bool,
   graph: ModuleGraph,
   loader: &'a mut dyn Loader,
   maybe_resolver: Option<&'a dyn Resolver>,
-  pending: FuturesUnordered<LoadFuture>,
+  pending: FuturesUnordered<LoadWithSpecifierFuture>,
   pending_assert_types: HashMap<ModuleSpecifier, HashSet<Option<String>>>,
   dynamic_branches: HashMap<ModuleSpecifier, Option<String>>,
   source_parser: &'a dyn SourceParser,
+  maybe_reporter: Option<&'a dyn Reporter>,
 }
 
 impl<'a> Builder<'a> {
@@ -1369,6 +1398,7 @@ impl<'a> Builder<'a> {
     maybe_resolver: Option<&'a dyn Resolver>,
     maybe_locker: Option<Rc<RefCell<Box<dyn Locker>>>>,
     source_parser: &'a dyn SourceParser,
+    maybe_reporter: Option<&'a dyn Reporter>,
   ) -> Self {
     Self {
       in_dynamic_branch: is_dynamic_root,
@@ -1379,11 +1409,13 @@ impl<'a> Builder<'a> {
       pending_assert_types: HashMap::new(),
       dynamic_branches: HashMap::new(),
       source_parser,
+      maybe_reporter,
     }
   }
 
   pub async fn build(
     mut self,
+    kind: BuildKind,
     maybe_imports: Option<Vec<(ModuleSpecifier, Vec<String>)>>,
   ) -> ModuleGraph {
     let roots = self.graph.roots.clone();
@@ -1414,30 +1446,40 @@ impl<'a> Builder<'a> {
     }
 
     loop {
-      match self.pending.next().await {
+      let specifier = match self.pending.next().await {
         Some((specifier, Ok(Some(response)))) => {
           let assert_types =
             self.pending_assert_types.remove(&specifier).unwrap();
           for maybe_assert_type in assert_types {
-            self.visit(&specifier, &response, maybe_assert_type)
+            self.visit(&specifier, &response, &kind, maybe_assert_type)
           }
+          Some(specifier)
         }
         Some((specifier, Ok(None))) => {
           self
             .graph
             .module_slots
-            .insert(specifier, ModuleSlot::Missing);
+            .insert(specifier.clone(), ModuleSlot::Missing);
+          Some(specifier)
         }
         Some((specifier, Err(err))) => {
           self.graph.module_slots.insert(
             specifier.clone(),
             ModuleSlot::Err(ModuleGraphError::LoadingErr(
-              specifier,
+              specifier.clone(),
               Arc::new(err),
             )),
           );
+          Some(specifier)
         }
-        _ => {}
+        _ => None,
+      };
+      if let (Some(specifier), Some(reporter)) =
+        (specifier, self.maybe_reporter)
+      {
+        let modules_total = self.graph.module_slots.len();
+        let modules_done = modules_total - self.pending.len();
+        reporter.on_load(&specifier, modules_done, modules_total);
       }
       if self.pending.is_empty() {
         // Start visiting queued up dynamic branches. We do this in a separate
@@ -1492,7 +1534,12 @@ impl<'a> Builder<'a> {
         .graph
         .module_slots
         .insert(specifier.clone(), ModuleSlot::Pending);
-      self.pending.push(self.loader.load(specifier, is_dynamic));
+      let specifier = specifier.clone();
+      let fut = self
+        .loader
+        .load(&specifier, is_dynamic)
+        .map(move |res| (specifier, res));
+      self.pending.push(Box::pin(fut));
     }
   }
 
@@ -1501,8 +1548,11 @@ impl<'a> Builder<'a> {
     &mut self,
     requested_specifier: &ModuleSpecifier,
     response: &LoadResponse,
+    kind: &BuildKind,
     maybe_assert_type: Option<String>,
   ) {
+    use std::borrow::BorrowMut;
+
     let maybe_headers = response.maybe_headers.as_ref();
     let specifier = response.specifier.clone();
 
@@ -1522,7 +1572,7 @@ impl<'a> Builder<'a> {
 
     let is_root = self.graph.roots.contains(&specifier);
 
-    let module_slot = parse_module(
+    let mut module_slot = parse_module(
       &specifier,
       maybe_headers,
       response.content.clone(),
@@ -1533,43 +1583,61 @@ impl<'a> Builder<'a> {
       self.in_dynamic_branch,
     );
 
-    if let ModuleSlot::Module(Module::Es(module)) = &module_slot {
-      for dep in module.dependencies.values() {
-        // Queue up dynamic dependencies to be visited later, unless we are
-        // already visiting a dynamic branch.
-        if dep.is_dynamic && !self.in_dynamic_branch {
-          if let Some(Ok((specifier, _))) = &dep.maybe_code {
-            self
-              .dynamic_branches
-              .insert(specifier.clone(), dep.maybe_assert_type.clone());
+    if let ModuleSlot::Module(Module::Es(module)) = module_slot.borrow_mut() {
+      if matches!(kind, BuildKind::All | BuildKind::CodeOnly)
+        || module.maybe_types_dependency.is_none()
+      {
+        for dep in module.dependencies.values_mut() {
+          if matches!(kind, BuildKind::All | BuildKind::CodeOnly)
+            || dep.maybe_type.is_none()
+          {
+            if let Some(Ok((specifier, _))) = &dep.maybe_code {
+              if dep.is_dynamic && !self.in_dynamic_branch {
+                self
+                  .dynamic_branches
+                  .insert(specifier.clone(), dep.maybe_assert_type.clone());
+              } else {
+                self.load(
+                  specifier,
+                  self.in_dynamic_branch,
+                  dep.maybe_assert_type.as_deref(),
+                );
+              }
+            }
+          } else {
+            dep.maybe_code = None;
           }
-          if let Some(Ok((specifier, _))) = &dep.maybe_type {
-            self
-              .dynamic_branches
-              .insert(specifier.clone(), dep.maybe_assert_type.clone());
-          }
-        } else {
-          if let Some(Ok((specifier, _))) = &dep.maybe_code {
-            self.load(
-              specifier,
-              self.in_dynamic_branch,
-              dep.maybe_assert_type.as_deref(),
-            );
-          }
-          if let Some(Ok((specifier, _))) = &dep.maybe_type {
-            self.load(
-              specifier,
-              self.in_dynamic_branch,
-              dep.maybe_assert_type.as_deref(),
-            );
+
+          if matches!(kind, BuildKind::All | BuildKind::TypesOnly) {
+            if let Some(Ok((specifier, _))) = &dep.maybe_type {
+              if dep.is_dynamic && !self.in_dynamic_branch {
+                self
+                  .dynamic_branches
+                  .insert(specifier.clone(), dep.maybe_assert_type.clone());
+              } else {
+                self.load(
+                  specifier,
+                  self.in_dynamic_branch,
+                  dep.maybe_assert_type.as_deref(),
+                );
+              }
+            }
+          } else {
+            dep.maybe_type = None;
           }
         }
+      } else {
+        module.dependencies.clear();
       }
 
-      if let Some((_, Some(Ok((specifier, _))))) =
-        &module.maybe_types_dependency
-      {
-        self.load(specifier, false, None);
+      if matches!(kind, BuildKind::All | BuildKind::TypesOnly) {
+        if let Some((_, Some(Ok((specifier, _))))) =
+          &module.maybe_types_dependency
+        {
+          self.load(specifier, false, None);
+        }
+      } else {
+        module.maybe_types_dependency = None;
       }
     }
     self.graph.module_slots.insert(specifier, module_slot);
@@ -1866,44 +1934,33 @@ mod tests {
             assert!(!is_dynamic);
             self.loaded_foo = true;
             Box::pin(async move {
-              (
-                specifier.clone(),
-                Ok(Some(LoadResponse {
-                  specifier: specifier.clone(),
-                  maybe_headers: None,
-                  content: Arc::new(
-                    "await import('file:///bar.js')".to_string(),
-                  ),
-                })),
-              )
+              Ok(Some(LoadResponse {
+                specifier: specifier.clone(),
+                maybe_headers: None,
+                content: Arc::new("await import('file:///bar.js')".to_string()),
+              }))
             })
           }
           "file:///bar.js" => {
             assert!(is_dynamic);
             self.loaded_bar = true;
             Box::pin(async move {
-              (
-                specifier.clone(),
-                Ok(Some(LoadResponse {
-                  specifier: specifier.clone(),
-                  maybe_headers: None,
-                  content: Arc::new("import 'file:///baz.js'".to_string()),
-                })),
-              )
+              Ok(Some(LoadResponse {
+                specifier: specifier.clone(),
+                maybe_headers: None,
+                content: Arc::new("import 'file:///baz.js'".to_string()),
+              }))
             })
           }
           "file:///baz.js" => {
             assert!(is_dynamic);
             self.loaded_baz = true;
             Box::pin(async move {
-              (
-                specifier.clone(),
-                Ok(Some(LoadResponse {
-                  specifier: specifier.clone(),
-                  maybe_headers: None,
-                  content: Arc::new("console.log('Hello, world!')".to_string()),
-                })),
-              )
+              Ok(Some(LoadResponse {
+                specifier: specifier.clone(),
+                maybe_headers: None,
+                content: Arc::new("console.log('Hello, world!')".to_string()),
+              }))
             })
           }
           _ => unreachable!(),
@@ -1923,8 +1980,9 @@ mod tests {
       None,
       None,
       &source_parser,
+      None,
     );
-    builder.build(None).await;
+    builder.build(BuildKind::All, None).await;
     assert!(loader.loaded_foo);
     assert!(loader.loaded_bar);
     assert!(loader.loaded_baz);
@@ -1942,18 +2000,13 @@ mod tests {
         let specifier = specifier.clone();
         match specifier.as_str() {
           "file:///foo.js" => Box::pin(async move {
-            (
-              specifier.clone(),
-              Ok(Some(LoadResponse {
-                specifier: specifier.clone(),
-                maybe_headers: None,
-                content: Arc::new("await import('file:///bar.js')".to_string()),
-              })),
-            )
+            Ok(Some(LoadResponse {
+              specifier: specifier.clone(),
+              maybe_headers: None,
+              content: Arc::new("await import('file:///bar.js')".to_string()),
+            }))
           }),
-          "file:///bar.js" => {
-            Box::pin(async move { (specifier.clone(), Ok(None)) })
-          }
+          "file:///bar.js" => Box::pin(async move { Ok(None) }),
           _ => unreachable!(),
         }
       }
@@ -1967,8 +2020,9 @@ mod tests {
       None,
       None,
       &source_parser,
+      None,
     );
-    let graph = builder.build(None).await;
+    let graph = builder.build(BuildKind::All, None).await;
     assert!(graph
       .try_get(&Url::parse("file:///foo.js").unwrap())
       .is_ok());
@@ -2006,24 +2060,18 @@ mod tests {
         let specifier = specifier.clone();
         match specifier.as_str() {
           "file:///foo.js" => Box::pin(async move {
-            (
-              specifier.clone(),
-              Ok(Some(LoadResponse {
-                specifier: Url::parse("file:///foo_actual.js").unwrap(),
-                maybe_headers: None,
-                content: Arc::new("import 'file:///bar.js'".to_string()),
-              })),
-            )
+            Ok(Some(LoadResponse {
+              specifier: Url::parse("file:///foo_actual.js").unwrap(),
+              maybe_headers: None,
+              content: Arc::new("import 'file:///bar.js'".to_string()),
+            }))
           }),
           "file:///bar.js" => Box::pin(async move {
-            (
-              specifier.clone(),
-              Ok(Some(LoadResponse {
-                specifier: Url::parse("file:///bar_actual.js").unwrap(),
-                maybe_headers: None,
-                content: Arc::new("(".to_string()),
-              })),
-            )
+            Ok(Some(LoadResponse {
+              specifier: Url::parse("file:///bar_actual.js").unwrap(),
+              maybe_headers: None,
+              content: Arc::new("(".to_string()),
+            }))
           }),
           _ => unreachable!(),
         }
@@ -2038,8 +2086,9 @@ mod tests {
       None,
       None,
       &source_parser,
+      None,
     );
-    let graph = builder.build(None).await;
+    let graph = builder.build(BuildKind::All, None).await;
     let specifiers = graph.specifiers();
     dbg!(&specifiers);
     assert_eq!(specifiers.len(), 4);
