@@ -430,6 +430,7 @@ mod tests {
   use super::*;
   use crate::ast::DefaultSourceParser;
   use crate::colors::strip_ansi_codes;
+  use crate::graph::BuildKind;
   use crate::graph::Builder;
   use crate::source::CacheInfo;
   use crate::source::MemoryLoader;
@@ -546,8 +547,9 @@ mod tests {
       None,
       None,
       &source_parser,
+      None,
     );
-    let graph = builder.build(None).await;
+    let graph = builder.build(BuildKind::All, None).await;
     assert_eq!(
       strip_ansi_codes(format!("{}", graph)),
       r#"local: /cache/deps/https/deno.land/x/example/a.ts
@@ -623,8 +625,9 @@ https://deno.land/x/example/a.ts (129B)
       None,
       None,
       &source_parser,
+      None,
     );
-    let graph = builder.build(None).await;
+    let graph = builder.build(BuildKind::All, None).await;
     println!("{}", graph);
     assert_eq!(
       strip_ansi_codes(format!("{}", graph)),
