@@ -11,6 +11,7 @@ mod module_specifier;
 pub mod source;
 mod text_encoding;
 
+use graph::BuildKind;
 use graph::Builder;
 use graph::ModuleSlot;
 use source::Locker;
@@ -46,9 +47,11 @@ cfg_if! {
     pub use module_specifier::SpecifierError;
 
     use source::Loader;
+    use source::Reporter;
 
     /// Create a module graph, based on loading and recursively analyzing the
     /// dependencies of the module, returning the resulting graph.
+    #[allow(clippy::too_many_arguments)]
     pub async fn create_graph(
       roots: Vec<ModuleSpecifier>,
       is_dynamic: bool,
@@ -57,6 +60,7 @@ cfg_if! {
       maybe_resolver: Option<&dyn Resolver>,
       maybe_locker: Option<Rc<RefCell<Box<dyn Locker>>>>,
       maybe_parser: Option<&dyn SourceParser>,
+      maybe_reporter: Option<&dyn Reporter>,
     ) -> ModuleGraph {
       let default_parser = ast::DefaultSourceParser::new();
       let source_parser = maybe_parser.unwrap_or(&default_parser);
@@ -67,8 +71,72 @@ cfg_if! {
         maybe_resolver,
         maybe_locker,
         source_parser,
+        maybe_reporter,
       );
-      builder.build(maybe_imports).await
+      builder.build(BuildKind::All, maybe_imports).await
+    }
+
+    /// Create a module graph, including only dependencies of the roots that
+    /// would contain code that would be executed, skipping any type only
+    /// dependencies. This is useful when wanting to build a graph of code for
+    /// loading in runtime that doesn't care about type only dependencies.
+    #[allow(clippy::too_many_arguments)]
+    pub async fn create_code_graph(
+      roots: Vec<ModuleSpecifier>,
+      is_dynamic: bool,
+      maybe_imports: Option<Vec<(ModuleSpecifier, Vec<String>)>>,
+      loader: &mut dyn Loader,
+      maybe_resolver: Option<&dyn Resolver>,
+      maybe_locker: Option<Rc<RefCell<Box<dyn Locker>>>>,
+      maybe_parser: Option<&dyn SourceParser>,
+      maybe_reporter: Option<&dyn Reporter>,
+    ) -> ModuleGraph {
+      let default_parser = ast::DefaultSourceParser::new();
+      let source_parser = maybe_parser.unwrap_or(&default_parser);
+      let builder = Builder::new(
+        roots,
+        is_dynamic,
+        loader,
+        maybe_resolver,
+        maybe_locker,
+        source_parser,
+        maybe_reporter,
+      );
+      builder.build(BuildKind::CodeOnly, maybe_imports).await
+    }
+
+    /// Create a module graph, including only dependencies that might affect
+    /// the types of the graph, skipping any "code only" imports. This is
+    /// useful in situations where only the types are being used, like when
+    /// type checking code or generating documentation.
+    ///
+    /// Note that code which is overloaded with types upon access (like the
+    /// `X-TypeScript-Types` header or types defined in the code itself) will
+    /// still be loaded into the graph, but further code only dependencies will
+    /// not be followed.
+    #[allow(clippy::too_many_arguments)]
+    pub async fn create_type_graph(
+      roots: Vec<ModuleSpecifier>,
+      is_dynamic: bool,
+      maybe_imports: Option<Vec<(ModuleSpecifier, Vec<String>)>>,
+      loader: &mut dyn Loader,
+      maybe_resolver: Option<&dyn Resolver>,
+      maybe_locker: Option<Rc<RefCell<Box<dyn Locker>>>>,
+      maybe_parser: Option<&dyn SourceParser>,
+      maybe_reporter: Option<&dyn Reporter>,
+    ) -> ModuleGraph {
+      let default_parser = ast::DefaultSourceParser::new();
+      let source_parser = maybe_parser.unwrap_or(&default_parser);
+      let builder = Builder::new(
+        roots,
+        is_dynamic,
+        loader,
+        maybe_resolver,
+        maybe_locker,
+        source_parser,
+        maybe_reporter,
+      );
+      builder.build(BuildKind::TypesOnly, maybe_imports).await
     }
 
     /// Parse an individual module, returning the module as a result, otherwise
@@ -138,6 +206,7 @@ cfg_if! {
       maybe_check: Option<js_sys::Function>,
       maybe_get_checksum: Option<js_sys::Function>,
       maybe_lockfile_name: Option<String>,
+      maybe_build_kind: Option<String>,
       maybe_imports: JsValue,
     ) -> Result<js_graph::ModuleGraph, JsValue> {
       let roots_vec: Vec<String> = roots.into_serde().map_err(|err| JsValue::from(js_sys::Error::new(&err.to_string())))?;
@@ -161,6 +230,11 @@ cfg_if! {
           .map_err(|err| JsValue::from(js_sys::Error::new(&err.to_string())))?;
         roots.push(root);
       }
+      let build_kind = match maybe_build_kind.as_deref() {
+        Some("typesOnly") => BuildKind::TypesOnly,
+        Some("codeOnly") => BuildKind::CodeOnly,
+        _ => BuildKind::All,
+      };
       let mut maybe_imports = None;
       if let Some(imports_map) = maybe_imports_map {
         let mut imports = Vec::new();
@@ -180,8 +254,9 @@ cfg_if! {
         maybe_resolver.as_ref().map(|r| r as &dyn Resolver),
         maybe_locker,
         &source_parser,
+        None,
       );
-      let graph = builder.build(maybe_imports).await;
+      let graph = builder.build(build_kind, maybe_imports).await;
       Ok(js_graph::ModuleGraph(graph))
     }
 
@@ -275,6 +350,7 @@ mod tests {
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(graph.module_slots.len(), 2);
@@ -342,9 +418,17 @@ mod tests {
       ModuleSpecifier::parse("file:///a/test01.ts").unwrap(),
       ModuleSpecifier::parse("https://example.com/a.ts").unwrap(),
     ];
-    let graph =
-      create_graph(roots.clone(), false, None, &mut loader, None, None, None)
-        .await;
+    let graph = create_graph(
+      roots.clone(),
+      false,
+      None,
+      &mut loader,
+      None,
+      None,
+      None,
+      None,
+    )
+    .await;
     assert_eq!(graph.module_slots.len(), 4);
     assert_eq!(graph.roots, roots);
     assert!(
@@ -369,9 +453,17 @@ mod tests {
       vec![],
     );
     let roots = vec![ModuleSpecifier::parse("file:///a/test.json").unwrap()];
-    let graph =
-      create_graph(roots.clone(), true, None, &mut loader, None, None, None)
-        .await;
+    let graph = create_graph(
+      roots.clone(),
+      true,
+      None,
+      &mut loader,
+      None,
+      None,
+      None,
+      None,
+    )
+    .await;
     assert_eq!(
       json!(graph),
       json!({
@@ -412,9 +504,17 @@ mod tests {
       vec![],
     );
     let roots = vec![ModuleSpecifier::parse("file:///a/test.js").unwrap()];
-    let graph =
-      create_graph(roots.clone(), true, None, &mut loader, None, None, None)
-        .await;
+    let graph = create_graph(
+      roots.clone(),
+      true,
+      None,
+      &mut loader,
+      None,
+      None,
+      None,
+      None,
+    )
+    .await;
     assert_eq!(
       json!(graph),
       json!({
@@ -490,6 +590,7 @@ console.log(a);
       None,
       None,
       None,
+      None,
     )
     .await;
     assert!(graph.valid().is_ok());
@@ -522,6 +623,7 @@ console.log(a);
       None,
       None,
       None,
+      None,
     )
     .await;
     assert!(graph.valid().is_err());
@@ -568,6 +670,7 @@ console.log(a);
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(
@@ -606,7 +709,7 @@ console.log(a);
             "dependencies": [
               {
                 "specifier": "./types_01.d.ts",
-                "code": {
+                "type": {
                   "specifier": "file:///a/types_01.d.ts",
                   "span": {
                     "start": {
@@ -682,6 +785,7 @@ console.log(a);
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(
@@ -731,6 +835,7 @@ console.log(a);
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(graph.module_slots.len(), 1);
@@ -783,6 +888,7 @@ console.log(a);
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(
@@ -848,6 +954,7 @@ console.log(a);
       None,
       None,
       None,
+      None,
     )
     .await;
     let result = graph.valid();
@@ -882,6 +989,7 @@ console.log(a);
       None,
       None,
       None,
+      None,
     )
     .await;
     let result = graph.valid();
@@ -922,6 +1030,7 @@ console.log(a);
       None,
       None,
       None,
+      None,
     )
     .await;
     assert!(graph.valid().is_ok());
@@ -952,6 +1061,7 @@ console.log(a);
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(
@@ -1009,7 +1119,7 @@ console.log(a);
             r#"
 /**
  * Some js doc
- * 
+ *
  * @param {import("./types.d.ts").A} a
  * @return {import("./other.ts").B}
  */
@@ -1043,6 +1153,7 @@ export function a(a) {
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(
@@ -1094,7 +1205,7 @@ export function a(a) {
               }
             ],
             "mediaType": "JavaScript",
-            "size": 138,
+            "size": 137,
             "specifier": "file:///a/test.js"
           },
           {
@@ -1142,6 +1253,7 @@ export function a(a) {
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(
@@ -1204,6 +1316,7 @@ export function a(a) {
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(
@@ -1261,6 +1374,7 @@ export function a(a) {
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(graph.module_slots.len(), 3);
@@ -1306,6 +1420,7 @@ export function a(a) {
       maybe_resolver,
       None,
       None,
+      None,
     )
     .await;
     let maybe_module = graph.get(&graph.roots[0]);
@@ -1366,6 +1481,7 @@ export function a(a) {
       maybe_resolver,
       None,
       None,
+      None,
     )
     .await;
     let maybe_module = graph.get(&graph.roots[0]);
@@ -1427,6 +1543,7 @@ export function a(a) {
       None,
       None,
       Some(&parser),
+      None,
     )
     .await;
     let root_ast = parser.get_parsed_source(&root_specifier).unwrap();
@@ -1486,6 +1603,7 @@ export function a(a) {
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(
@@ -1629,6 +1747,7 @@ export function a(a) {
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(
@@ -1721,6 +1840,7 @@ export function a(a) {
       None,
       None,
       None,
+      None,
     )
     .await;
     assert_eq!(
@@ -1847,6 +1967,509 @@ export function a(a) {
     );
   }
 
+  #[derive(Debug)]
+  struct CollectingReporter {
+    on_loads: RefCell<Vec<(ModuleSpecifier, usize, usize)>>,
+  }
+
+  impl source::Reporter for CollectingReporter {
+    fn on_load(
+      &self,
+      specifier: &ModuleSpecifier,
+      modules_done: usize,
+      modules_total: usize,
+    ) {
+      self.on_loads.borrow_mut().push((
+        specifier.clone(),
+        modules_done,
+        modules_total,
+      ));
+    }
+  }
+
+  #[tokio::test]
+  async fn test_create_graph_with_reporter() {
+    let mut loader = setup(
+      vec![
+        (
+          "file:///a/test01.ts",
+          Ok((
+            "file:///a/test01.ts",
+            None,
+            r#"
+            import "./a.js";
+            "#,
+          )),
+        ),
+        (
+          "file:///a/a.js",
+          Ok(("file:///a/a.js", None, r#"import "./b.js";"#)),
+        ),
+        (
+          "file:///a/b.js",
+          Ok((
+            "file:///a/b.js",
+            None,
+            r#"import "./c.js"; import "./d.js""#,
+          )),
+        ),
+        ("file:///a/c.js", Ok(("file:///a/c.js", None, r#""#))),
+        ("file:///a/d.js", Ok(("file:///a/d.js", None, r#""#))),
+      ],
+      vec![],
+    );
+    let root_specifier =
+      ModuleSpecifier::parse("file:///a/test01.ts").expect("bad url");
+    let reporter = CollectingReporter {
+      on_loads: RefCell::new(vec![]),
+    };
+    let graph = create_graph(
+      vec![root_specifier.clone()],
+      false,
+      None,
+      &mut loader,
+      None,
+      None,
+      None,
+      Some(&reporter),
+    )
+    .await;
+    assert_eq!(graph.modules().len(), 5);
+
+    let on_loads = reporter.on_loads.into_inner();
+    assert_eq!(on_loads.len(), 5);
+
+    assert_eq!(
+      on_loads[0],
+      (ModuleSpecifier::parse("file:///a/test01.ts").unwrap(), 1, 2)
+    );
+    assert_eq!(
+      on_loads[1],
+      (ModuleSpecifier::parse("file:///a/a.js").unwrap(), 2, 3)
+    );
+    assert_eq!(
+      on_loads[2],
+      (ModuleSpecifier::parse("file:///a/b.js").unwrap(), 3, 5)
+    );
+    let c = ModuleSpecifier::parse("file:///a/c.js").unwrap();
+    if on_loads[3].0 == c {
+      assert_eq!(
+        on_loads[3],
+        (ModuleSpecifier::parse("file:///a/c.js").unwrap(), 4, 5)
+      );
+      assert_eq!(
+        on_loads[4],
+        (ModuleSpecifier::parse("file:///a/d.js").unwrap(), 5, 5)
+      );
+    } else {
+      assert_eq!(
+        on_loads[3],
+        (ModuleSpecifier::parse("file:///a/d.js").unwrap(), 4, 5)
+      );
+      assert_eq!(
+        on_loads[4],
+        (ModuleSpecifier::parse("file:///a/c.js").unwrap(), 5, 5)
+      );
+    }
+  }
+
+  #[tokio::test]
+  async fn test_create_type_graph() {
+    let mut loader = setup(
+      vec![
+        (
+          "file:///a/test01.ts",
+          Ok((
+            "file:///a/test01.ts",
+            None,
+            r#"
+            // @deno-types="./a.d.ts"
+            import * as a from "./a.js";
+            import type { B } from "./b.d.ts";
+            import * as c from "https://example.com/c";
+            import * as d from "./d.js";
+            "#,
+          )),
+        ),
+        (
+          "file:///a/a.js",
+          Ok(("file:///a/a.js", None, r#"export const a = "a""#)),
+        ),
+        (
+          "file:///a/a.d.ts",
+          Ok(("file:///a/a.d.ts", None, r#"export const a: "a";"#)),
+        ),
+        (
+          "file:///a/b.d.ts",
+          Ok(("file:///a/b.d.ts", None, r#"export interface B {}"#)),
+        ),
+        (
+          "https://example.com/c",
+          Ok((
+            "https://example.com/c",
+            Some(vec![
+              ("x-typescript-types", "./c.d.ts"),
+              ("content-type", "application/javascript"),
+            ]),
+            r#"export { c } from "./c.js";"#,
+          )),
+        ),
+        (
+          "https://example.com/c.d.ts",
+          Ok((
+            "https://example.com/c.d.ts",
+            Some(vec![("content-type", "application/typescript")]),
+            r#"export const c: "c";"#,
+          )),
+        ),
+        (
+          "https://example.com/c.js",
+          Ok((
+            "https://example.com/c.js",
+            Some(vec![("content-type", "application/javascript")]),
+            r#"export const c = "c";"#,
+          )),
+        ),
+        (
+          "file:///a/d.js",
+          Ok(("file:///a/d.js", None, r#"export const d = "d";"#)),
+        ),
+      ],
+      vec![],
+    );
+    let root_specifier =
+      ModuleSpecifier::parse("file:///a/test01.ts").expect("bad url");
+    let graph = create_type_graph(
+      vec![root_specifier.clone()],
+      false,
+      None,
+      &mut loader,
+      None,
+      None,
+      None,
+      None,
+    )
+    .await;
+    assert_eq!(
+      json!(graph),
+      json!({
+        "roots": [
+          "file:///a/test01.ts"
+        ],
+        "modules": [
+          {
+            "dependencies": [],
+            "mediaType": "Dts",
+            "size": 20,
+            "specifier": "file:///a/a.d.ts"
+          },
+          {
+            "dependencies": [],
+            "mediaType": "Dts",
+            "size": 21,
+            "specifier": "file:///a/b.d.ts"
+          },
+          {
+            "dependencies": [],
+            "mediaType": "JavaScript",
+            "size": 21,
+            "specifier": "file:///a/d.js"
+          },
+          {
+            "dependencies": [
+              {
+                "specifier": "./a.js",
+                "type": {
+                  "specifier": "file:///a/a.d.ts",
+                  "span": {
+                    "start": {
+                      "line": 1,
+                      "character": 28
+                    },
+                    "end": {
+                      "line": 1,
+                      "character": 36
+                    }
+                  }
+                }
+              },
+              {
+                "specifier": "./b.d.ts",
+                "type": {
+                  "specifier": "file:///a/b.d.ts",
+                  "span": {
+                    "start": {
+                      "line": 3,
+                      "character": 35
+                    },
+                    "end": {
+                      "line": 3,
+                      "character": 45
+                    }
+                  }
+                }
+              },
+              {
+                "specifier": "./d.js",
+                "code": {
+                  "specifier": "file:///a/d.js",
+                  "span": {
+                    "start": {
+                      "line": 5,
+                      "character": 31
+                    },
+                    "end": {
+                      "line": 5,
+                      "character": 39
+                    }
+                  }
+                }
+              },
+              {
+                "specifier": "https://example.com/c",
+                "code": {
+                  "specifier": "https://example.com/c",
+                  "span": {
+                    "start": {
+                      "line": 4,
+                      "character": 31
+                    },
+                    "end": {
+                      "line": 4,
+                      "character": 54
+                    }
+                  }
+                }
+              }
+            ],
+            "mediaType": "TypeScript",
+            "size": 236,
+            "specifier": "file:///a/test01.ts"
+          },
+          {
+            "dependencies": [],
+            "typesDependency": {
+              "specifier": "./c.d.ts",
+              "dependency": {
+                "specifier": "https://example.com/c.d.ts",
+                "span": {
+                  "start": {
+                    "line": 0,
+                    "character": 0
+                  },
+                  "end": {
+                    "line": 0,
+                    "character": 0
+                  }
+                }
+              }
+            },
+            "mediaType": "JavaScript",
+            "size": 27,
+            "specifier": "https://example.com/c"
+          },
+          {
+            "dependencies": [],
+            "mediaType": "Dts",
+            "size": 20,
+            "specifier": "https://example.com/c.d.ts"
+          }
+        ],
+        "redirects": {}
+      })
+    );
+  }
+
+  #[tokio::test]
+  async fn test_create_code_graph() {
+    let mut loader = setup(
+      vec![
+        (
+          "file:///a/test01.ts",
+          Ok((
+            "file:///a/test01.ts",
+            None,
+            r#"
+            // @deno-types="./a.d.ts"
+            import * as a from "./a.js";
+            import type { B } from "./b.d.ts";
+            import * as c from "https://example.com/c";
+            import * as d from "./d.js";
+            "#,
+          )),
+        ),
+        (
+          "file:///a/a.js",
+          Ok(("file:///a/a.js", None, r#"export const a = "a""#)),
+        ),
+        (
+          "file:///a/a.d.ts",
+          Ok(("file:///a/a.d.ts", None, r#"export const a: "a";"#)),
+        ),
+        (
+          "file:///a/b.d.ts",
+          Ok(("file:///a/b.d.ts", None, r#"export interface B {}"#)),
+        ),
+        (
+          "https://example.com/c",
+          Ok((
+            "https://example.com/c",
+            Some(vec![
+              ("x-typescript-types", "./c.d.ts"),
+              ("content-type", "application/javascript"),
+            ]),
+            r#"export { c } from "./c.js";"#,
+          )),
+        ),
+        (
+          "https://example.com/c.d.ts",
+          Ok((
+            "https://example.com/c.d.ts",
+            Some(vec![("content-type", "application/typescript")]),
+            r#"export const c: "c";"#,
+          )),
+        ),
+        (
+          "https://example.com/c.js",
+          Ok((
+            "https://example.com/c.js",
+            Some(vec![("content-type", "application/javascript")]),
+            r#"export const c = "c";"#,
+          )),
+        ),
+        (
+          "file:///a/d.js",
+          Ok(("file:///a/d.js", None, r#"export const d = "d";"#)),
+        ),
+      ],
+      vec![],
+    );
+    let root_specifier =
+      ModuleSpecifier::parse("file:///a/test01.ts").expect("bad url");
+    let graph = create_code_graph(
+      vec![root_specifier.clone()],
+      false,
+      None,
+      &mut loader,
+      None,
+      None,
+      None,
+      None,
+    )
+    .await;
+    // println!("{}", serde_json::to_string_pretty(&graph).unwrap());
+    assert_eq!(
+      json!(graph),
+      json!({
+        "roots": [
+          "file:///a/test01.ts"
+        ],
+        "modules": [
+          {
+            "dependencies": [],
+            "mediaType": "JavaScript",
+            "size": 20,
+            "specifier": "file:///a/a.js"
+          },
+          {
+            "dependencies": [],
+            "mediaType": "JavaScript",
+            "size": 21,
+            "specifier": "file:///a/d.js"
+          },
+          {
+            "dependencies": [
+              {
+                "specifier": "./a.js",
+                "code": {
+                  "specifier": "file:///a/a.js",
+                  "span": {
+                    "start": {
+                      "line": 2,
+                      "character": 31
+                    },
+                    "end": {
+                      "line": 2,
+                      "character": 39
+                    }
+                  }
+                }
+              },
+              {
+                "specifier": "./b.d.ts",
+              },
+              {
+                "specifier": "./d.js",
+                "code": {
+                  "specifier": "file:///a/d.js",
+                  "span": {
+                    "start": {
+                      "line": 5,
+                      "character": 31
+                    },
+                    "end": {
+                      "line": 5,
+                      "character": 39
+                    }
+                  }
+                }
+              },
+              {
+                "specifier": "https://example.com/c",
+                "code": {
+                  "specifier": "https://example.com/c",
+                  "span": {
+                    "start": {
+                      "line": 4,
+                      "character": 31
+                    },
+                    "end": {
+                      "line": 4,
+                      "character": 54
+                    }
+                  }
+                }
+              }
+            ],
+            "mediaType": "TypeScript",
+            "size": 236,
+            "specifier": "file:///a/test01.ts"
+          },
+          {
+            "dependencies": [
+              {
+                "specifier": "./c.js",
+                "code": {
+                  "specifier": "https://example.com/c.js",
+                  "span": {
+                    "start": {
+                      "line": 0,
+                      "character": 18
+                    },
+                    "end": {
+                      "line": 0,
+                      "character": 26
+                    }
+                  }
+                }
+              }
+            ],
+            "mediaType": "JavaScript",
+            "size": 27,
+            "specifier": "https://example.com/c"
+          },
+          {
+            "dependencies": [],
+            "mediaType": "JavaScript",
+            "size": 21,
+            "specifier": "https://example.com/c.js"
+          }
+        ],
+        "redirects": {}
+      })
+    );
+  }
+
   #[test]
   fn test_parse_module() {
     let specifier = ModuleSpecifier::parse("file:///a/test01.ts").unwrap();
@@ -2017,7 +2640,7 @@ export function a(a) {
         r#"
 /**
  * Some js doc
- * 
+ *
  * @param {import("./types.d.ts").A} a
  * @return {import("./other.ts").B}
  */
@@ -2070,7 +2693,7 @@ export function a(a) {
           }
         ],
         "mediaType": "JavaScript",
-        "size": 138,
+        "size": 137,
         "specifier": "file:///a/test.js"
       })
     );
