@@ -54,7 +54,7 @@ pub struct LoadResponse {
   pub content: Arc<String>,
 }
 
-pub type LoadResult = (ModuleSpecifier, Result<Option<LoadResponse>>);
+pub type LoadResult = Result<Option<LoadResponse>>;
 pub type LoadFuture = Pin<Box<dyn Future<Output = LoadResult> + 'static>>;
 
 /// A trait which allows asynchronous loading of source files into a module
@@ -123,7 +123,7 @@ pub trait Resolver: fmt::Debug {
   }
 }
 
-pub(crate) fn load_data_url(
+pub fn load_data_url(
   specifier: &ModuleSpecifier,
 ) -> Result<Option<LoadResponse>> {
   let url = DataUrl::process(specifier.as_str())
@@ -205,10 +205,26 @@ impl Loader for MemoryLoader {
       None if specifier.scheme() == "data" => load_data_url(specifier),
       _ => Ok(None),
     };
-    Box::pin(future::ready((specifier.clone(), response)))
+    Box::pin(future::ready(response))
   }
 }
 
+/// A trait which can be used to allow the module graph to report status events
+/// to the user.
+pub trait Reporter: fmt::Debug {
+  /// A handler that is called after each load of a module. It contains the
+  /// module specifier of the module that was loaded, and the number of modules
+  /// seen (total number of unique specifiers seen), and the number of modules
+  /// loaded (where [Loader::load] has been called, and the returned future is
+  /// ready).
+  fn on_load(
+    &self,
+    specifier: &ModuleSpecifier,
+    modules_done: usize,
+    modules_total: usize,
+  );
+}
+
 #[cfg(test)]
 pub mod tests {
   use super::*;
